import { sync } from 'most-subject';
import invariant from 'invariant';
import $$observable from 'symbol-observable';
import { combine, from } from 'most';

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function step(key, arg) {
        try {
          var info = gen[key](arg);
          var value = info.value;
        } catch (error) {
          reject(error);
          return;
        }

        if (info.done) {
          resolve(value);
        } else {
          Promise.resolve(value).then(_next, _throw);
        }
      }

      function _next(value) {
        step("next", value);
      }

      function _throw(err) {
        step("throw", err);
      }

      _next();
    });
  };
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function warning(message) {
  if (typeof console !== 'undefined' && typeof console.error === 'function') {
    console.error(message);
  }
  try {
    throw new Error(message);
  } catch (e) {
  }
}

const counter = () => {
  let id = 0;
  return () => ++id;
};
function toTag(...tags) {
  return tags.filter(str => str.length > 0).join('/');
}
const nextPayloadID = counter();
const nextEventID = counter();

const PING = '@@effector/ping';
const PONG = '@@effector/pong';
const HALT = '@@effector/halt';
function createHaltAction() {
  return {
    type: HALT,
    payload: undefined
  };
}

function isAction(data) {
  return typeof data === 'object' && data != null && typeof data.type === 'string';
}
function safeDispatch(data, dispatch) {
  if (Array.isArray(data)) {
    data.filter(isAction).forEach(action => dispatch(action));
  } else if (isAction(data)) dispatch(data);
  return data;
}
function port(dispatch, state$, action$) {
  return function epic(handler) {
    const result = handler(action$, state$).multicast();
    result.observe(data => safeDispatch(data, dispatch));
    return result;
  };
}

class Scope$1 {
  constructor() {
    this.dispatchRef = new WeakMap();
    this.dispatchSet = new WeakMap();
  }
}

function genericSafeGet$1(store, instance, fabric) {
  const result = store.get(instance);
  if (result !== undefined) return result;
  const value = fabric();
  store.set(instance, value);
  return value;
}
function getFromSetMap$1(setMap, that) {
  return genericSafeGet$1(setMap, that, () => new Set());
}
function getFromWeakMap$1(setMap, that) {
  return genericSafeGet$1(setMap, that, () => new WeakMap());
}

function get(ref, instance, eventName) {
  const weakRef = getFromWeakMap$1(ref, instance);
  return getFromSetMap$1(weakRef, eventName);
}
function add(value, ref, instance, eventName) {
  get(ref, instance, eventName).add(value);
}
function remove(value, ref, instance, eventName) {
  const weakRef = ref.get(instance);
  if (!weakRef) return;
  const set = weakRef.get(eventName);
  if (!set) return;
  set.delete(value);
}
function size(ref, instance, eventName) {
  const weakRef = ref.get(instance);
  if (!weakRef) return 0;
  const set = weakRef.get(eventName);
  if (!set) return 0;
  return set.size;
}
function clear(ref, instance, eventName) {
  const weakRef = ref.get(instance);
  if (!weakRef) return;
  const set = weakRef.get(eventName);
  if (!set) return;
  return set.clear();
}

function get$1(ref, eventName) {
  return getFromSetMap$1(ref, eventName);
}
function add$1(ref, instance, eventName) {
  get$1(ref, eventName).add(instance);
}
function remove$1(ref, instance, eventName) {
  const eventRef = ref.get(eventName);
  if (!eventRef) {
    return;
  }
  eventRef.delete(instance);
}
function clear$1(ref, instance, eventName) {
  const eventRef = ref.get(eventName);
  if (!eventRef) {
    return;
  }
  return eventRef.clear();
}

function dispatch(scope, event, value) {
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;
  try {
    for (var _iterator = get$1(scope.dispatchSet, event)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      const instance = _step.value;
      invariant(instance !== event, 'circular reference');
      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;
      try {
        for (var _iterator2 = get(scope.dispatchRef, instance, event)[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          const fn = _step2.value;
          fn(value);
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
            _iterator2.return();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }
}

let id$1 = 0;
class Key {
  constructor() {
    this.id = ++id$1;
  }
  toString() {
    return this.id.toString(36);
  }
  inspect() {
    return this.toString();
  }
}
function createKey() {
  return new Key();
}

function add$2(scope, instance, event, fn) {
  add$1(scope.dispatchSet, instance, event);
  add(fn, scope.dispatchRef, instance, event);
}
function remove$2(scope, instance, event, fn) {
  remove(fn, scope.dispatchRef, instance, event);
  const refSize = size(scope.dispatchRef, instance, event);
  if (refSize > 0) return;
  remove$1(scope.dispatchSet, instance, event);
}
function clear$2(scope, instance, event) {
  clear(scope.dispatchRef, instance, event);
  clear$1(scope.dispatchSet, instance, event);
}

const defaultScope$1 = new Scope$1();
class Emission {
  constructor() {
    this.key = createKey();
  }
  watch(event, fn, scope = defaultScope$1) {
    add$2(scope, this.key, event.key, fn);
    return () => remove$2(scope, this.key, event.key, fn);
  }
  dispatch(value, scope = defaultScope$1) {
    dispatch(scope, this.key, value);
  }
  unlink(event, scope = defaultScope$1) {
    const key = event instanceof Emission ? event.key : event;
    clear$2(scope, this.key, key);
  }
  clear(scope = defaultScope$1) {
    const set = get$1(scope.dispatchSet, this.key);
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;
    try {
      for (var _iterator = set[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        const event = _step.value;
        this.unlink(event, scope);
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return != null) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }
    set.clear();
  }
}

function basicCommon(domainName, name) {
  const type = toTag(domainName, name);
  return {
    eventID: nextEventID(),
    nextSeq: counter(),
    getType: () => type
  };
}
function observable(event, action$) {
  function observableState() {
    return {
      subscribe(observer) {
        invariant(typeof observer === 'object', 'Expected the observer to be an object');
        return action$.subscribe(observer);
      },
      [O]() {
        return event;
      }
    };
  }
  const O = $$observable;
  event[O] = observableState;
}

function watchWarn(domainName, name, error) {
  warning(`


  Watch function should not fail

  Effect:
    ${name}
  Domain:
    ${domainName}
  Error:
    ${error}

`);
}
function implWarn(domainName, name, props) {
  warning(`


  Running an effect without implementation

  Name:
    ${name}
  Domain:
    ${domainName}
  Arguments:
    ${props}


`);
}
function implError(domainName, name, props) {
  const message = `Running an effect without implementation

  Name:
    ${name}
  Domain:
    ${domainName}
  Arguments:
    ${props}


`;
  return new Error(message);
}

function createWatcher({
  event,
  domainName,
  name,
  opts
}) {
  return function watch(fn) {
    event.epic((data$, state$) => combine((data, state) => ({
      data,
      state
    }), data$, state$).sampleWith(data$).map(({
      data,
      state
    }) => {
      try {
        const result = fn(data, state);
        return result;
      } catch (err) {
        watchFailWarn({
          domainName,
          name,
          mode: opts.watchFailCheck()
        }, err);
      }
    }));
  };
}
function watchFailWarn({
  domainName,
  name,
  mode
}, error) {
  switch (mode) {
    case 'throw':
      {
        throw error;
      }
    case 'warn':
      {
        watchWarn(domainName, name, error);
      }
    case 'off':
    default:
      {
        return;
      }
  }
}

function EventConstructor(domainName, dispatch, getState, state$, events, name, action$ = sync(), config = {
  isPlain: false,
  watchFailCheck: 'warn'
}) {
  const _basicCommon = basicCommon(domainName, name),
        eventID = _basicCommon.eventID,
        nextSeq = _basicCommon.nextSeq,
        getType = _basicCommon.getType;
  const emission = new Emission();
  const handlers = new Set();
  if (!config.isPlain) handlers.add((payload
  ) => action$.next(payload));
  const create = payload => ({
    type: getType(),
    payload,
    meta: {
      index: nextPayloadID(),
      eventID,
      seq: nextSeq(),
      passed: false,
      plain: config.isPlain
    }
  });
  function eventInstance(payload) {
    const toSend = Promise.resolve(payload);
    const result = create(payload);
    let canDispatch = true;
    return _objectSpread({}, result, {
      raw() {
        return result;
      },
      send(dispatchHook) {
        if (canDispatch) {
          const dispatcher = dispatchHook ? dispatchHook : dispatch;
          canDispatch = false;
          dispatcher(result);
          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;
          try {
            for (var _iterator = handlers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              const handler = _step.value;
              handler(payload, getState());
            }
          } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion && _iterator.return != null) {
                _iterator.return();
              }
            } finally {
              if (_didIteratorError) {
                throw _iteratorError;
              }
            }
          }
          emission.dispatch(result);
        }
        return toSend;
      }
    });
  }
  eventInstance.emission = emission;
  eventInstance.getType = getType;
  eventInstance.toString = getType;
  eventInstance.watch = createWatcher({
    event: eventInstance,
    domainName,
    name,
    opts: {
      watchFailCheck: () => config.watchFailCheck
    }
  });
  eventInstance.epic = port(dispatch, state$, action$);
  eventInstance.subscribe = subscriber => action$.subscribe(subscriber);
  eventInstance.trigger = function trigger(query, eventName = 'trigger') {
    const triggerEvent = EventConstructor(domainName, dispatch, getState, state$, events, [name, eventName].join(' '));
    triggerEvent.watch((_, state) => {
      const queryResult = query(state);
      return eventInstance(queryResult);
    });
    return triggerEvent;
  };
  observable(eventInstance, action$);
  events.set(getType(), eventInstance);
  return eventInstance;
}

function EffectConstructor(domainName, dispatch, getState, state$, events, name, opts, action$ = sync()) {
  const emission = new Emission();
  const handlers = new Set();
  const _basicCommon = basicCommon(domainName, name),
        eventID = _basicCommon.eventID,
        nextSeq = _basicCommon.nextSeq,
        getType = _basicCommon.getType;
  handlers.add((payload, state) => action$.next(payload));
  let using = getWarn({
    domainName,
    name,
    mode: () => opts.unused()
  });
  function use(thunk) {
    using = thunk;
  }
  const create = payload => ({
    type: getType(),
    payload,
    meta: {
      index: nextPayloadID(),
      eventID,
      seq: nextSeq(),
      passed: false
    }
  });
  const effect = params => {
    let run = () => {
      run = () => {};
      const runned = runUsing(params, using);
      runned.then(result => {
        const data = {
          params,
          result
        };
        dispatcher(done(data), dispatch);
        resolve(data);
      }, error => {
        const data = {
          params,
          error
        };
        dispatcher(fail(data), dispatch);
        reject(data);
      });
    };
    const runner = dispatchHook => {
      if (canDispatch) {
        canDispatch = false;
        dispatcher(result, dispatch, dispatchHook);
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;
        try {
          for (var _iterator = handlers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            const handler = _step.value;
            handler(params, getState());
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator.return != null) {
              _iterator.return();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }
        run();
      }
    };
    let resolve = val => {};
    let reject = val => {};
    const doneP = new Promise(rs => resolve = rs);
    const failP = new Promise(rs => reject = rs);
    const toSend = Promise.resolve(params);
    const result = create(params);
    let canDispatch = true;
    return _objectSpread({}, result, {
      raw() {
        return result;
      },
      send(dispatchHook) {
        runner(dispatchHook);
        return toSend;
      },
      done() {
        runner();
        return doneP;
      },
      fail() {
        runner();
        return failP;
      },
      promise() {
        runner();
        return Promise.race([doneP, failP.then(data => Promise.reject(data))]);
      }
    });
  };
  const done = EventConstructor(domainName, dispatch, getState, state$, events, [name, 'done'].join(' '));
  const fail = EventConstructor(domainName, dispatch, getState, state$, events, [name, 'fail'].join(' '));
  effect.use = use;
  effect.getType = getType;
  setToString(effect, getType);
  effect.done = done;
  effect.fail = fail;
  effect.epic = port(dispatch, state$, action$);
  effect.watch = createWatcher({
    event: effect,
    domainName,
    name,
    opts
  });
  effect.trigger = function trigger(query, eventName = 'trigger') {
    const triggerEvent = EventConstructor(domainName, dispatch, getState, state$, events, [name, eventName].join(' '));
    triggerEvent.watch((_, state) => {
      const queryResult = query(state);
      return effect(queryResult);
    });
    return triggerEvent;
  };
  effect.emission = emission;
  effect.subscribe = subscriber => action$.subscribe(subscriber);
  observable(effect, action$);
  return effect;
}
function setToString(effect, getType) {
  effect.toString = getType;
}
function runUsing(params, using) {
  let response;
  try {
    response = using(params);
  } catch (err) {
    return Promise.reject(err);
  }
  if (typeof response === 'object' && response != null && typeof response.then === 'function') {
    return response;
  }
  return Promise.resolve(response);
}
function dispatcher(value, dispatchDefault, dispatchHook, isBatch = false) {
  if (!isAction$1(value)) {
    return value;
  }
  const dispatch = typeof dispatchHook === 'function' ? dispatchHook : dispatchDefault;
  const result = dispatch(value);
  return result;
}
function isAction$1(data) {
  return typeof data === 'object' && data != null && typeof data.type === 'string';
}
const getWarn = ({
  domainName,
  name,
  mode
}) => function never(props) {
  switch (mode()) {
    case 'throw':
      {
        return Promise.reject(implError(domainName, name, props));
      }
    case 'warn':
      {
        implWarn(domainName, name, props);
      }
    case 'off':
    default:
      {
        return new Promise(() => {});
      }
  }
};

function createDomain(store, domainName = '') {
  const redux = store;
  const config = redux.dispatch({
    type: PING
  }).payload;
  const addPlainHandler = (name, fn) => {
    config.plain.set(name, fn);
  };
  const state$ = from(redux).map(() => store.getState()).multicast();
  return DomainConstructor(domainName, redux.dispatch, store.getState, state$, new Map(), addPlainHandler);
}
function createRootDomain(domainName = '') {
  const addPlainHandler = (name, fn) => {
    config.plain.set(name, fn);
  };
  let redux;
  let reduxStore;
  const config = {
    plain: new Map()
  };
  const dispatch = function dispatch(data) {
    return redux.dispatch(data);
  };
  function getState() {
    return reduxStore.getState();
  }
  const state$ = sync();
  let unsubscribe;
  function disablePrevious(previousDispatch) {
    previousDispatch(createHaltAction());
    if (typeof unsubscribe === 'function') {
      unsubscribe();
    }
  }
  function register(store) {
    const previousDispatch = redux != null ? redux.dispatch : v => v;
    redux = reduxStore = store;
    const pong = redux.dispatch({
      type: PING,
      payload: undefined
    });
    if (pong) {
      const payload = pong.payload;
      const plain = payload.plain;
      const oldPlain = config.plain;
      oldPlain.clear();
      config.plain = plain;
    }
    disablePrevious(previousDispatch);
    const unsub = from(store).map(() => getState()).subscribe({
      next(value) {
        state$.next(value);
      },
      error(err) {},
      complete() {}
    });
    unsubscribe = () => unsub.unsubscribe();
  }
  const domain = DomainConstructor(domainName, dispatch, getState, state$, new Map(), addPlainHandler);
  domain.register = register;
  return domain;
}
function optsReducer([key, value]) {
  if (!(key in this)) return;
  const newVal = () => value;
  this[key] = newVal;
}
function mergeEffectOpts(defaults, opts) {
  if (opts === undefined) return _objectSpread({}, defaults);
  invariant(typeof opts === 'object' && opts != null, 'effect options should be object, got %s', typeof opts);
  const resultObj = _objectSpread({}, defaults);
  Object.keys(opts).map(key => [key, opts[key]]).forEach(optsReducer, resultObj);
  return resultObj;
}
function DomainConstructor(domainName, dispatch, getState, state$, events = new Map(), addPlainHandler = () => {}, configGetter) {
  let getConfig;
  if (typeof configGetter !== 'function') {
    const config = {
      unused: () => 'warn',
      watchFailCheck: () => 'warn',
      dispatch
    };
    getConfig = () => config;
  } else getConfig = configGetter;
  const disp = getDispatch(getConfig);
  const effectOpts = {
    unused() {
      return getConfig().unused();
    },
    watchFailCheck() {
      return getConfig().watchFailCheck();
    },
    dispatch: disp
  };
  function recoverable(events$) {
    const str$ = events$.multicast();
    const failureHandler = e => {
      portFailWarn({
        domainName
      }, e);
      return str$;
    };
    return str$.recoverWith(failureHandler);
  }
  return {
    port(events$) {
      return recoverable(events$).observe(data => {
        safeDispatch(data, dispatch);
      });
    },
    register(store) {
      warning(`Not implemented`);
    },
    effect(name, opts) {
      const optsFull = mergeEffectOpts(effectOpts, opts);
      return EffectConstructor(domainName, dispatch, getState, state$, events, name, optsFull);
    },
    domain(name) {
      return DomainConstructor(toTag(domainName, name), dispatch, getState, state$, events, addPlainHandler, getConfig);
    },
    event(name) {
      return EventConstructor(domainName, dispatch, getState, state$, events, name);
    },
    typeConstant(name) {
      const action$ = sync();
      addPlainHandler(name, data => {
        if (data.meta == null) data.meta = {};
        data.meta.plain = true;
        const resultEvent = result(data.payload);
        resultEvent.meta.passed = true;
        resultEvent.meta.plain = true;
        action$.next(data.payload);
        getConfig().dispatch(resultEvent);
      });
      const result = EventConstructor('', dispatch, getState, state$, events, name, action$, {
        isPlain: true,
        watchFailCheck: effectOpts.watchFailCheck()
      });
      return result;
    }
  };
}
function getDispatch(getConfig) {
  const dispatch = value => {
    const result = getConfig().dispatch(value);
    return result;
  };
  return dispatch;
}
function portFailWarn({
  domainName
}, error) {
  warning(`


  Port stream should not fail.
  Next failure will throw.

  Domain:
    ${domainName}
  Error:
    ${error}

`);
}

function effectorMiddleware() {
  const plainMap = new Map();
  const config = {
    plain: plainMap
  };
  let isActive = true;
  return next => action => {
    if (!isActive) {
      return action;
    }
    if (!(typeof action === 'object' && action != null)) {
      return next(action);
    }
    if (typeof action.type !== 'string') {
      return action;
    }
    if (action.type === PING) {
      return {
        type: PONG,
        payload: config
      };
    }
    if (action.type === HALT) {
      isActive = false;
      plainMap.clear();
      return next(action);
    }
    const handler = plainMap.get(action.type);
    if (!(typeof action.meta === 'object' && action.meta != null)) action.meta = {};
    if (action.meta.passed === true) return action;
    if (handler) action.meta.plain = true;
    if (typeof action.send === 'function') {
      const send = action.send,
            meta = action.meta;
      if (handler) {
        meta.passed = true;
        send(next);
        handler(action);
        return action;
      }
      action.send(next);
      return action;
    } else {
      const meta = action.meta;
      if (handler) {
        if (meta.passed !== true) {
          meta.passed = true;
          const result = next(action);
          handler(action);
          return result;
        }
        action.meta.passed = true;
      }
      action.meta.passed = true;
      return next(action);
    }
  };
}

function atom(defaultValue) {
  return new Atomic(defaultValue);
}
function remote(baseValue) {
  let value = baseValue;
  return {
    get() {
      return value();
    },
    set(newValue) {
      value = () => newValue;
    }
  };
}
function createAtom(ref) {
  const active = remote(() => true);
  const emission = new Emission();
  const set = new Emission();
  const get = new Emission();
  const offA = emission.watch(set, ref.set);
  const offB = emission.watch(get, cb => cb(ref.get()));
  const off = () => {
    active.set(false);
    offA();
    offB();
  };
  return {
    get,
    set,
    off,
    active,
    emission
  };
}
function getReturn(defaultValue, get) {
  let value;
  get.dispatch(newValue => {
    value = newValue;
  });
  if (value == null) return defaultValue();
  return value;
}
class Atomic {
  constructor(defaultValue) {
    this.meta = void 0;
    let fnValue;
    const defValue = () => fnValue();
    if (defaultValue instanceof Atomic) {
      fnValue = () => {
        const result = defaultValue.get();
        fnValue = () => result;
        return result;
      };
    } else {
      fnValue = () => defaultValue;
    }
    const _createAtom = createAtom(remote(defValue)),
          get = _createAtom.get,
          set = _createAtom.set,
          off = _createAtom.off,
          active = _createAtom.active,
          emission = _createAtom.emission;
    Object.defineProperty(this, 'meta', {
      value() {
        return {
          get,
          set,
          off,
          active,
          emission,
          defaultValue: defValue
        };
      }
    });
  }
  get() {
    const _meta = this.meta(),
          active = _meta.active,
          defaultValue = _meta.defaultValue,
          get = _meta.get;
    invariant(active.get(), 'inactive atom');
    return getReturn(defaultValue, get);
  }
  set(newValue) {
    if (newValue == null) return;
    const _meta2 = this.meta(),
          active = _meta2.active,
          set = _meta2.set;
    invariant(active.get(), 'inactive atom');
    const currentState = this.get();
    if (currentState === newValue) return;
    set.dispatch(newValue);
  }
  isActive() {
    return this.meta().active.get();
  }
  off() {
    this.meta().off();
  }
  watch(fn) {
    const _meta3 = this.meta(),
          emission = _meta3.emission,
          set = _meta3.set;
    return emission.watch(set, fn);
  }
  map(fn) {
    return new Reference(_ => [fn(_)], this);
  }
  chain(fn) {
    return new Reference(fn, this);
  }
  sample(b, fn) {
    let bVal = b.get();
    const offB = b.watch(b => {
      bVal = b;
    });
    const offA = iterate(this, {
      next(val) {
        const result = fn(val, bVal);
        atom.set(result);
      },
      complete() {
        atom.off();
        offA();
        offB();
      }
    });
    const atom = new Atomic();
    return atom;
  }
  combine(b, fn) {
    let r;
    const set = (a, b) => {
      const result = fn(a, b);
      if (result == null) return;
      if (r === result) return;
      r = result;
      atom.set(result);
    };
    const offB = b.watch(b => {
      set(this.get(), b);
    });
    const offA = iterate(this, {
      next(val) {
        set(val, b.get());
      },
      complete() {
        atom.off();
        offA();
        offB();
      }
    });
    const atom = new Atomic();
    return atom;
  }
  update(fn) {
    const oldValue = this.get();
    const newValue = fn(oldValue);
    if (oldValue === newValue) return this;
    if (newValue == null) return this;
    this.set(newValue);
    return this;
  }
  reset() {
    return this.set(this.meta().defaultValue());
  }
}
function iterate(ref, observable) {
  const off = ref.watch(_ => observable.next(_));
  return () => {
    observable.complete();
    off();
  };
}
class Reference extends Atomic {
  constructor(transformer, ref) {
    const defs = [...transformer(ref.get())];
    super(defs[defs.length - 1]);
    const halt = () => this.off();
    const set = t => this.set(t);
    const off = iterate(ref, {
      next(val) {
        var _iteratorNormalCompletion3 = true;
        var _didIteratorError3 = false;
        var _iteratorError3 = undefined;
        try {
          for (var _iterator3 = transformer(val)[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
            const result = _step3.value;
            set(result);
          }
        } catch (err) {
          _didIteratorError3 = true;
          _iteratorError3 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion3 && _iterator3.return != null) {
              _iterator3.return();
            }
          } finally {
            if (_didIteratorError3) {
              throw _iteratorError3;
            }
          }
        }
      },
      complete() {
        halt();
        off();
      }
    });
  }
}

function normalizeType(typeOrActionCreator) {
  if (typeOrActionCreator && typeOrActionCreator.getType) {
    return typeOrActionCreator.toString();
  }
  return typeOrActionCreator;
}
function createReducer(defaultState, handlers = {}, thisType) {
  const lastState = atom(defaultState);
  return ReducerConstructor(lastState, handlers, thisType);
}
function ReducerConstructor(lastState, handlers, thisType) {
  const opts = {
    fallback: null,
    defaultState: lastState.get()
  };
  const reducer = Object.assign(reduce, {
    has,
    map,
    watch,
    get,
    set,
    on,
    off,
    options,
    reset,
    lastState,
    subscribe
  });
  const returnType = thisType ? thisType : reducer;
  function watch(fn) {
    return lastState.watch(fn);
  }
  function subscribe(subscriber) {
    invariant(typeof subscriber.next === 'function', 'wrong subscriber');
    return {
      unsubscribe: lastState.watch(_ => subscriber.next(_))
    };
  }
  function map(fn) {
    const state = lastState.map(fn);
    return ReducerConstructor(state, {});
  }
  function get() {
    return lastState.get();
  }
  function set(value) {
    lastState.update(lastState => {
      if (value == null) return;
      if (lastState === value) return;
      return value;
    });
    return reducer;
  }
  function has(typeOrActionCreator) {
    return !!handlers[normalizeType(typeOrActionCreator)];
  }
  function reset(typeOrActionCreator) {
    on(typeOrActionCreator, (state, payload, meta) => opts.defaultState);
    lastState.reset();
    return returnType;
  }
  function on(typeOrActionCreator, handler) {
    if (Array.isArray(typeOrActionCreator)) {
      typeOrActionCreator.forEach(action => {
        on(action, handler);
      });
    } else {
      handlers[normalizeType(typeOrActionCreator)] = handler;
    }
    return returnType;
  }
  function off(typeOrActionCreator) {
    if (Array.isArray(typeOrActionCreator)) {
      typeOrActionCreator.forEach(off);
    } else {
      delete handlers[normalizeType(typeOrActionCreator)];
    }
    return returnType;
  }
  function options(newOpts) {
    Object.keys(newOpts).forEach(name => opts[name] = newOpts[name]);
    return returnType;
  }
  if (typeof handlers === 'function') {
    const factory = handlers;
    handlers = {};
    factory(on, off);
  }
  function reduce(stateRaw, action) {
    const state = stateRaw === undefined ? opts.defaultState : stateRaw;
    if (!action || typeof action.type !== 'string') {
      set(state);
      return state;
    }
    if (action.type.startsWith('@@redux/')) {
      set(state);
      return state;
    }
    const handler = handlers[String(action.type)] || opts.fallback;
    if (handler) {
      const result = handler(state, action.payload, action.meta);
      if (result == null) {
        return get();
      }
      set(result);
      return result;
    }
    set(state);
    return state;
  }
  return reducer;
}

function combine$1(combine$$1, ...reducers) {
  const values = reducers.map(reducer => reducer(undefined, {
    type: '@@effector/void'
  }));
  const stateAtom = atom(combine$$1(...values));
  const resultStates = values.map(v => atom(v));
  for (let i = 0; i < resultStates.length; i++) {
    const off = resultStates[i].watch(update => {
      const prevState = stateAtom.get();
      const updated = combine$$1(...resultStates.map(s => s.get()));
      if (updated !== prevState) stateAtom.set(updated);
    });
  }
  const defaultState = combine$$1(...values);
  const combined = createReducer(stateAtom.get(), {}, combinedReducer);
  function combinedReducer(state, action) {
    for (let i = 0; i < reducers.length; i++) {
      const statePart = resultStates[i];
      const reducer = reducers[i];
      const prevValue = statePart.get();
      const value = reducer(prevValue, action);
      if (value !== prevValue) {
        statePart.set(value);
      }
    }
    return stateAtom.get();
  }
  Object.assign(combinedReducer, combined);
  return combinedReducer;
}

let defaultDomain;
function domainGetter() {
  if (defaultDomain === undefined) {
    const res = createRootDomain();
    defaultDomain = res;
  }
  return defaultDomain;
}
function createEvent$1(name) {
  const domain = domainGetter();
  const event = domain.event(name);
  return event;
}
function createEffect(name) {
  const domain = domainGetter();
  const effect = domain.effect(name);
  return effect;
}

class Collect {
  and(red) {
    return new Collect([...this.parts, red]);
  }
  combine(fn) {
    const reducers = this.parts.length === 0 ? [createReducer()] : this.parts;
    const join = combine$1;
    const joined = join(fn, ...reducers);
    return joined;
  }
  joint(fn) {
    return this.combine(fn);
  }
  constructor(parts = []) {
    this.parts = void 0;
    this.parts = parts;
  }
}
function collect() {
  return new Collect();
}

const INIT = '@@redux/INIT' + Math.random().toString(36).substring(7).split('').join('.');
const REPLACE = '@@redux/REPLACE' + Math.random().toString(36).substring(7).split('').join('.');

function compose(...funcs) {
  if (funcs.length === 0) {
    return arg => arg;
  }
  if (funcs.length === 1) {
    return funcs[0];
  }
  return funcs.reduce((a, b) => (...args) => a(b(...args)));
}

function applyMiddleware(...middlewares) {
  return createStore => (...args) => {
    const store = createStore(...args);
    let _dispatch = () => {
      throw new Error(`Dispatching while constructing your middleware is not allowed. ` + `Other middleware would not be applied to this dispatch.`);
    };
    let chain = [];
    const middlewareAPI = {
      getState: store.getState,
      dispatch: (...args) => _dispatch(...args)
    };
    chain = middlewares.map(middleware => middleware(middlewareAPI));
    _dispatch = compose(...chain)(store.dispatch);
    return _objectSpread({}, store, {
      dispatch: _dispatch
    });
  };
}

function storeConstructor(props) {
  let currentListeners = [];
  let isDispatching = false;
  let currentReducer = props.currentReducer,
      currentState = props.currentState;
  let nextListeners = currentListeners;
  function ensureCanMutateNextListeners() {
    if (nextListeners === currentListeners) {
      nextListeners = currentListeners.slice();
    }
  }
  function getState() {
    invariant(!isDispatching, 'You may not call store.getState() while the reducer is executing. ' + 'The reducer has already received the state as an argument. ' + 'Pass it down from the top reducer instead of reading it from the store.');
    return currentState;
  }
  function subscribe(listener) {
    invariant(typeof listener === 'function', 'Expected the listener to be a function.');
    invariant(!isDispatching, 'You may not call store.subscribe() while the reducer is executing. ' + 'If you would like to be notified after the store has been updated, subscribe from a ' + 'component and invoke store.getState() in the callback to access the latest state');
    let isSubscribed = true;
    ensureCanMutateNextListeners();
    nextListeners.push(listener);
    return function unsubscribe() {
      if (!isSubscribed) {
        return;
      }
      invariant(!isDispatching, 'You may not unsubscribe from a store listener while the reducer is executing');
      isSubscribed = false;
      ensureCanMutateNextListeners();
      const index = nextListeners.indexOf(listener);
      nextListeners.splice(index, 1);
    };
  }
  function dispatch(action) {
    invariant(typeof action.type !== 'undefined', 'Actions may not have an undefined "type" property.');
    invariant(!isDispatching, 'Reducers may not dispatch actions.');
    try {
      isDispatching = true;
      currentState = currentReducer(currentState, action);
    } finally {
      isDispatching = false;
    }
    const listeners = currentListeners = nextListeners;
    for (let i = 0; i < listeners.length; i++) {
      const listener = listeners[i];
      listener();
    }
    return action;
  }
  function replaceReducer(nextReducer) {
    invariant(typeof nextReducer === 'function', 'Expected the nextReducer to be a function.');
    currentReducer = nextReducer;
    dispatch({
      type: REPLACE
    });
  }
  function observable() {
    const outerSubscribe = subscribe;
    return {
      subscribe(observer) {
        invariant(typeof observer === 'object' && observer != null, 'Expected the observer to be an object.');
        function observeState() {
          if (observer.next) {
            observer.next(getState());
          }
        }
        observeState();
        const unsubscribe = outerSubscribe(observeState);
        return {
          unsubscribe
        };
      },
      [$$observable]() {
        return this;
      }
    };
  }
  dispatch({
    type: INIT
  });
  return {
    dispatch,
    subscribe,
    getState,
    replaceReducer,
    [$$observable]: observable
  };
}
function createStore(reducer, preloadedStateRaw, enhancerRaw) {
  let enhancer = enhancerRaw;
  let preloadedState = preloadedStateRaw;
  if (typeof preloadedState === 'function' && typeof enhancer === 'undefined') {
    enhancer = preloadedState;
    preloadedState = undefined;
  }
  if (typeof enhancer !== 'undefined') {
    if (typeof enhancer !== 'function') {
      invariant(Array.isArray(enhancer), 'Expected the enhancer to be a function');
      enhancer = applyMiddleware(...enhancer);
    }
    return enhancer(createStore)(reducer, preloadedState);
  }
  invariant(typeof reducer === 'function', 'Expected the reducer to be a function');
  return storeConstructor({
    currentReducer: reducer,
    currentState: preloadedState
  });
}

class PlainActorSystem {
  constructor() {
    this.isDispatching = false;
    this.pending = new Set();
  }
}
const defaultActorSystem = new PlainActorSystem();

function seq(system = defaultActorSystem) {
  if (system.isDispatching) return;
  system.isDispatching = true;
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;
  try {
    for (var _iterator = forPendingActors(system.pending)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      const _ = _step.value;
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator.return != null) {
        _iterator.return();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }
  system.isDispatching = false;
}
function* forPendingActors(pending) {
  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;
  try {
    for (var _iterator2 = untilEnd(pending)[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      const _ref = _step2.value;
      const inbox = _ref.inbox,
            reducer = _ref.reducer,
            state = _ref.state;
      yield* forActorEvents(inbox, reducer, state);
    }
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }
}
function* forActorEvents(inbox, reducer, state) {
  var _iteratorNormalCompletion3 = true;
  var _didIteratorError3 = false;
  var _iteratorError3 = undefined;
  try {
    for (var _iterator3 = untilEnd(inbox)[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
      const event = _step3.value;
      yield* forReducer(event, reducer, state);
    }
  } catch (err) {
    _didIteratorError3 = true;
    _iteratorError3 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion3 && _iterator3.return != null) {
        _iterator3.return();
      }
    } finally {
      if (_didIteratorError3) {
        throw _iteratorError3;
      }
    }
  }
}
function* forReducer(event, reducer, state) {
  var _iteratorNormalCompletion4 = true;
  var _didIteratorError4 = false;
  var _iteratorError4 = undefined;
  try {
    for (var _iterator4 = reducer[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
      const fn = _step4.value;
      yield fn(state.get(), event, state.set);
    }
  } catch (err) {
    _didIteratorError4 = true;
    _iteratorError4 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion4 && _iterator4.return != null) {
        _iterator4.return();
      }
    } finally {
      if (_didIteratorError4) {
        throw _iteratorError4;
      }
    }
  }
}
function* untilEnd(set) {
  do {
    var _arr = [...set];
    for (var _i = 0; _i < _arr.length; _i++) {
      const e = _arr[_i];
      set.delete(e);
      yield e;
    }
  } while (set.size > 0);
}

class Quant {
  constructor(value = null) {
    this.id = ++Quant.id;
    this.value = void 0;
    this.value = value;
    Object.defineProperties(this, {
      get: {
        value: this.get.bind(this),
        configurable: true,
        writable: true
      },
      set: {
        value: this.set.bind(this),
        configurable: true,
        writable: true
      }
    });
  }
  get() {
    return this.value;
  }
  set(value) {
    if (value === this) return;
    this.value = value;
  }
  map(fn) {
    return new Quant(fn(this.get()));
  }
}
Quant.id = 0;
class OpaqueEvent {
  constructor(type, payload, meta) {
    this.id = ++OpaqueEvent.id;
    this.type = void 0;
    this.payload = void 0;
    this.meta = void 0;
    this.type = type;
    this.payload = payload;
    this.meta = meta || {};
  }
  valueOf() {
    return {
      id: this.id,
      payload: this.payload.get(),
      type: this.type
    };
  }
  raw() {
    return this.valueOf();
  }
  send() {}
}
OpaqueEvent.id = 0;
class EventFabric {
  constructor(type) {
    this.id = ++EventFabric.id;
    this.type = void 0;
    this.subscribers = new Set();
    this.type = type;
  }
  create(payload, meta, system = defaultActorSystem) {
    const box = new Quant(payload);
    const event = new OpaqueEvent(this.type, box, meta);
    invariant(event.payload, 'no payload');
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;
    try {
      for (var _iterator = this.subscribers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        const actor = _step.value;
        dispatch$1(event, actor, system);
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return != null) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }
    seq(system);
    return event;
  }
  add(actor) {
    this.subscribers.add(actor);
  }
}
EventFabric.id = 0;
class PlainActor {
  constructor(state, reducer = new Set(), inbox = new Set())
  {
    this.state = void 0;
    this.inbox = void 0;
    this.reducer = void 0;
    const newState = state instanceof Quant ? state : new Quant(state);
    this.state = newState;
    this.reducer = reducer;
    this.inbox = inbox;
  }
  getState() {
    return this.state.get();
  }
  replaceReducer(fn) {
    this.reducer = fn(this.reducer);
  }
  addReducer(reducer) {
    this.replaceReducer(_ => _.add(reducer));
  }
}
function dispatch$1(event, actor, system = defaultActorSystem) {
  system.pending.add(actor);
  actor.inbox.add(event);
}

class Event extends OpaqueEvent {
  constructor(type, value) {
    const data = value instanceof Atom ? value : new Atom(value);
    super(type, data);
    this.payload = void 0;
  }
}
class Atom extends Quant {
  constructor(value) {
    super(value);
  }
  *[Symbol.iterator]() {
    yield* this.iter();
  }
  get() {
    return super.get();
  }
  set(value) {
    return super.set(value);
  }
}
class EventFabric$1 extends EventFabric {
  constructor(type) {
    super(type);
    this.Event = void 0;
    class TypedEvent extends Event {}
    this.Event = TypedEvent;
  }
  create(payload, meta, system = defaultActorSystem) {
    const Event = this.Event;
    const event = new Event(this.type, payload);
    event.payload = new Atom(payload);
    invariant(event.payload, 'no payload');
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;
    try {
      for (var _iterator = this.subscribers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        const actor = _step.value;
        dispatch$1(event, actor, system);
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return != null) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }
    seq(system);
    return event;
  }
  is(event) {
    return event instanceof this.Event;
  }
}
function eventFabric(type) {
  return new EventFabric$1(type);
}
class Actor extends PlainActor {
  constructor(state, reducer = new Set(), inbox = new Set())
  {
    super(state);
    this.state = void 0;
    this.inbox = void 0;
    this.reducer = void 0;
    this.state = state;
    this.reducer = reducer;
    this.inbox = inbox;
  }
  getState() {
    return this.state.get();
  }
  replaceReducer(fn) {
    this.reducer = fn(this.reducer);
  }
  addReducer(reducer) {
    this.replaceReducer(_ => _.add(reducer));
  }
  on(eventCreator, handler) {
    const fabric = eventCreator instanceof EventFabric$1 ? [eventCreator] : [...eventCreator];
    for (var _i = 0; _i < fabric.length; _i++) {
      const fab = fabric[_i];
      fab.add(this);
    }
    const isEvent = _ => fabric.some(fab => fab.is(_));
    this.addReducer((state, event, setState) => {
      if (isEvent(event)) {
        handler(state, event, setState);
      }
    });
    return this;
  }
  raise(value, fabric, actor) {
    dispatch$1(fabric.create(value), actor);
  }
}
function createActorState(value) {
  return new Actor(new Atom(value));
}

class MutableQueue {
  constructor() {
    this.size = 0;
    this.head = null;
    this.last = null;
  }
  tag() {
    return 'MutableQueue';
  }
}
class PutQueue extends MutableQueue {
  constructor(...args) {
    var _temp;
    return _temp = super(...args), this.head = void 0, this.last = void 0, _temp;
  }
}
class DelegateQueue extends MutableQueue {
  constructor(...args) {
    var _temp2;
    return _temp2 = super(...args), this.head = void 0, this.last = void 0, _temp2;
  }
}
class MutableCell {
  constructor(queue, value) {
    this.value = void 0;
    this.queue = void 0;
    this.next = null;
    this.prev = null;
    this.queue = queue;
    this.value = value;
  }
  tag() {
    return 'MutableCell';
  }
}

function writeAVar(avar, value) {
  avar.value = value;
  return avar;
}
function erase(avar) {
  return writeAVar(avar, null);
}
function putLast(queue, value) {
  const cell = new MutableCell(queue, value);
  switch (queue.size) {
    case 0:
      queue.head = cell;
      break;
    case 1:
      {
        const head = queue.head;
        invariant(head, 'no head');
        cell.prev = head;
        head.next = cell;
        queue.last = cell;
        break;
      }
    default:
      {
        const last = queue.last;
        invariant(last, 'no last');
        cell.prev = last;
        last.next = cell;
        queue.last = cell;
      }
  }
  queue.size++;
  return cell;
}
function takeLast(queue) {
  let cell;
  switch (queue.size) {
    case 0:
      return null;
    case 1:
      cell = queue.head;
      queue.head = null;
      break;
    case 2:
      {
        const head = queue.head;
        invariant(head, 'no head');
        cell = queue.last;
        head.next = null;
        queue.last = null;
        break;
      }
    default:
      {
        const last = queue.last;
        invariant(last, 'no last');
        cell = last;
        const prev = last.prev;
        invariant(prev, 'no prev');
        prev.next = null;
        queue.last = prev;
      }
  }
  invariant(cell, 'no cell');
  cell.prev = null;
  cell.queue = null;
  queue.size--;
  return cell.value;
}
function takeHead(queue) {
  let cell;
  switch (queue.size) {
    case 0:
      return null;
    case 1:
      cell = queue.head;
      queue.head = null;
      break;
    case 2:
      {
        const last = queue.last;
        invariant(last, 'no last');
        cell = queue.head;
        last.prev = null;
        queue.head = last;
        queue.last = null;
        break;
      }
    default:
      {
        cell = queue.head;
        invariant(cell, 'no cell');
        const _cell = cell,
              next = _cell.next;
        invariant(next, 'no next');
        next.prev = null;
        queue.head = next;
      }
  }
  invariant(cell, 'no cell');
  cell.next = null;
  cell.queue = null;
  queue.size--;
  return cell.value;
}
function deleteCell(cell) {
  if (cell.queue == null) {
    return;
  }
  if (cell.queue.last === cell) {
    takeLast(cell.queue);
    return;
  }
  if (cell.queue.head === cell) {
    takeHead(cell.queue);
    return;
  }
  if (cell.prev) {
    cell.prev.next = cell.next;
  }
  if (cell.next) {
    cell.next.prev = cell.prev;
  }
  cell.queue.size--;
  cell.queue = null;
  cell.value = null;
  cell.next = null;
  cell.prev = null;
}
function drainVar(avar) {
  if (avar.draining) {
    return;
  }
  const puts = avar.puts,
        takes = avar.takes,
        reads = avar.reads;
  let p;
  let r;
  let t;
  let value;
  let rsize;
  avar.draining = true;
  while (1) {
    p = null;
    r = null;
    t = null;
    value = avar.value;
    rsize = reads.size;
    if (avar.error != null) {
      value = avar.error;
      while (p = takeHead(puts)) {
        runEff(value, p.next, p.error);
      }
      while (r = takeHead(reads)) {
        runEff(value, r);
      }
      while (t = takeHead(takes)) {
        runEff(value, t);
      }
      break;
    }
    if ((null === value || undefined === value) && (p = takeHead(puts))) {
      writeAVar(avar, value = p.value);
    }
    if (null !== value && undefined !== value) {
      t = takeHead(takes);
      while (rsize-- && (r = takeHead(reads))) {
        runEff(value, r);
      }
      if (null !== t && undefined !== t) {
        erase(avar);
        runEff(value, t);
      }
    }
    if (null !== p && undefined !== p) {
      runEff(void 0, p.next, p.error);
    }
    if ((null === avar.value || undefined === avar.value) && puts.size === 0 || null !== avar.value && undefined !== avar.value && takes.size === 0) {
      break;
    }
  }
  avar.draining = false;
}
function drainErase(avar) {
  drainVar(erase(avar));
}
function runEff(value, eff, onError) {
  try {
    eff(value);
  } catch (error) {
    (onError || doThrow)(error);
  }
}
function doThrow(error) {
  setTimeout(() => {
    throw error;
  }, 0);
}

function _killVar(error, avar) {
  if (avar.error == null) {
    avar.error = error;
    drainErase(avar);
  }
}
function _putVar(value, avar, cb, error) {
  const cell = putLast(avar.puts, {
    next: cb,
    value,
    error
  });
  drainVar(avar);
  return () => {
    deleteCell(cell);
  };
}
function _tryPutVar(value, avar) {
  if ((null === avar.value || undefined === avar.value) && avar.error == null) {
    drainVar(writeAVar(avar, value));
    return true;
  }
  return false;
}
function asyncPutVar(value, avar) {
  return new Promise(_ => {
    _putVar(value, avar, _);
  });
}
function _takeVar(avar, cb) {
  const cell = putLast(avar.takes, cb);
  drainVar(avar);
  return () => {
    deleteCell(cell);
  };
}
function _tryTakeVar(avar) {
  const value = avar.value;
  if (null === value || undefined === value) {
    return null;
  }
  drainErase(avar);
  return value;
}
function asyncTakeVar(avar) {
  return new Promise(_ => {
    _takeVar(avar, _);
  });
}
function _readVar(avar, cb) {
  const cell = putLast(avar.reads, cb);
  drainVar(avar);
  return () => {
    deleteCell(cell);
  };
}
function asyncReadVar(avar) {
  return new Promise(_ => {
    _readVar(avar, _);
  });
}
function _tryReadVar(avar) {
  if (null === avar.value || undefined === avar.value) {
    return null;
  }
  return avar.value;
}

class AVar {
  constructor(value) {
    this.draining = false;
    this.error = null;
    this.takes = new DelegateQueue();
    this.reads = new DelegateQueue();
    this.puts = new PutQueue();
    this.value = void 0;
    this.value = value === undefined ? null : value;
  }
  read(cb) {
    return _readVar(this, cb);
  }
  asyncRead() {
    return asyncReadVar(this);
  }
  put(value, cb, error) {
    return _putVar(value, this, cb, error);
  }
  asyncReplace(value) {
    if (this.status().filled) {
      this.tryTake();
    }
    this.tryPut(value);
    return Promise.resolve();
  }
  asyncPut(value) {
    return asyncPutVar(value, this);
  }
  take(cb) {
    return _takeVar(this, cb);
  }
  asyncTake() {
    return asyncTakeVar(this);
  }
  tryRead() {
    return _tryReadVar(this);
  }
  tryTake() {
    return _tryTakeVar(this);
  }
  tryPut(value) {
    invariant(value !== undefined && value !== null, 'no value');
    return _tryPutVar(value, this);
  }
  status() {
    const isEmpty = this.value === undefined || this.value === null;
    const killed = this.error != null;
    return {
      empty: isEmpty,
      filled: !isEmpty,
      killed
    };
  }
  kill(error) {
    _killVar(error, this);
  }
  tag() {
    return 'AVar';
  }
}
function makeVar(value) {
  return new AVar(value);
}

function syncSetter(reducer, state, act, avar, setState) {
  const result = reducer(state, act, _ => {
    if (avar.status().filled) avar.tryTake();
    setState(_);
  });
  console.log(result);
}
function asyncSetter(_x, _x2, _x3, _x4) {
  return _asyncSetter.apply(this, arguments);
}
function _asyncSetter() {
  _asyncSetter = _asyncToGenerator(function* (reducer, act, avar, setState) {
    const state = yield avar.asyncTake();
    const payload = yield new Promise(_ => reducer(state, act, val => (setState(val), _(val))));
  });
  return _asyncSetter.apply(this, arguments);
}
class ActorVar {
  constructor(value) {
    this.avar = makeVar();
    this.actor = createActorState(this.avar);
    if (value !== undefined) this.avar.tryPut(value);
  }
  on(event, reducer) {
    this.actor.on(event, (state, data, setState) => {
      const setter = _ => {
        state.asyncPut(_);
      };
      const stateValue = state.tryTake();
      const act = data.raw();
      if (stateValue !== null && stateValue !== undefined) {
        return syncSetter(reducer, stateValue, act, state, setter);
      }
      asyncSetter(reducer, act, state, setter);
    });
    return this;
  }
  status() {
    return this.avar.status();
  }
  take() {
    return this.avar.asyncTake();
  }
  read() {
    return this.avar.asyncRead();
  }
  put(value) {
    return this.avar.asyncPut(value);
  }
  readSync() {
    return this.avar.tryRead();
  }
  takeSync() {
    return this.avar.tryTake();
  }
  putSync(value) {
    return this.avar.tryPut(value);
  }
}
function createActor$1(value) {
  return new ActorVar(value);
}

export { createDomain, createRootDomain, effectorMiddleware, createReducer, createStore, combine$1 as combine, combine$1 as joint, createHaltAction, createEvent$1 as createEvent, createEffect, collect as mill, collect, Atomic, atom, Reference, eventFabric as event, createActor$1 as createActor, ActorVar, EventFabric$1 as EventFabric, Event as EventNew };
//# sourceMappingURL=effector.es.js.map
